import datetime
import os.path
import sys

import click

from . import image_fetching
from . import vmt
from . import vtf

#TODO: Refactor to eliminate repeated code
@click.group()
def cli():
    pass

def get_module_directory() -> str:
    module_path = os.path.abspath(__file__)
    module_dir = os.path.dirname(module_path)
    return module_dir
@cli.command(help = "Set TF2 directory.")
@click.argument("tf2_dir", type = click.Path(exists = True))
def set_directory(tf2_dir):
    module_dir = get_module_directory()
    tf2_dir_file = os.path.join(module_dir, "TF2_DIRECTORY.txt")
    with open(tf2_dir_file, "wt") as file:
        file.write(tf2_dir)

#TODO: ensure TF2_DIRECTORY.txt is always in the same folder as the script for neatness (default working directory not always proper with vscode)
def get_tf2_directory() -> str:
    try: 
        module_dir = get_module_directory()
        tf2_dir_file = os.path.join(module_dir, "TF2_DIRECTORY.txt")
        with open (tf2_dir_file, "rt") as file:
            return file.read()
    except FileNotFoundError:
        raise FileNotFoundError("TF2 directory not configured! Please run pyspray set-directory before using the script!")


source_argument = click.argument("source", type = str)
format_option = click.option("-f", "--format", type = click.Choice([format.name for format in vtf.ImageFormats if format.implemented], case_sensitive = False), default = "DXT5", help = "The format in which the spray is encoded.")
aspect_ratio_flag = click.option("--preserve_aspect_ratio/--no_preserve_aspect_ratio", is_flag = True, default = True, help = "Whether the aspect ratio of the image should be preserved. If true, the image will be padded to square using either transparent or black padding, determined by the image format's alpha support.")
name_option = click.option("-n", "--name", type = str, help = "The name of the spray.")

@cli.command(help = "Create an animated spray. Argument is a filepath or a url. Local filepaths are checked before attempting url retrieval.")
@source_argument
@format_option
@aspect_ratio_flag
@name_option
@click.option("-start", "--start_time", type = float, default = 0, help = "The time in seconds at which the animation starts.")
@click.option("-end", "--end_time", type = float, default = sys.float_info.max, help = "The time in seconds at which the animation ends.")
@click.option("-spf", "--seconds_per_frame", type = float, default = 0.2, help = "The number of seconds in between each frame of the animation. The default value is the same framerate as natively possessed by sprays, so any change will speed up or slow down the animation.")
def anim(source, format, preserve_aspect_ratio, name, start_time, end_time, seconds_per_frame):
    tf2_directory = get_tf2_directory()
    images = image_fetching.get_images(source, seconds_per_frame, start_time, end_time)
    flags =     vtf.TextureFlags.CLAMPS |\
                vtf.TextureFlags.CLAMPT |\
                vtf.TextureFlags.NOLOD |\
                vtf.TextureFlags.EIGHTBITALPHA |\
                vtf.TextureFlags.NOMIP 
    animated_spray = vtf.AnimatedSpray(images, vtf.ImageFormats.from_name(format), flags, preserve_aspect_ratio)
    spray_path = os.path.join(tf2_directory, "tf", "materials", "vgui", "logos", f"{name}.vtf")
    animated_spray.save(spray_path)
    vmt.write_vmt_files(name, tf2_directory)

@cli.command(help = "Create a spray that changes with distance. Arguments are filepaths or urls. Filepaths are checked before attempting url retrieval. The first source will be shown from furthest away, and the last source will be shown when you get closest to the spray." )
@click.argument("source", nargs = -1, type = str)
@format_option
@aspect_ratio_flag
@name_option
def fade(source, format, preserve_aspect_ratio, name):
    tf2_directory = get_tf2_directory()
    images = [image_fetching.get_images(path)[0] for path in source]
    name = name or datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

    flags =     vtf.TextureFlags.CLAMPS |\
                vtf.TextureFlags.CLAMPT |\
                vtf.TextureFlags.NOLOD |\
                vtf.TextureFlags.EIGHTBITALPHA
    fade_spray = vtf.FadeSpray(images, vtf.ImageFormats.from_name(format), flags, preserve_aspect_ratio)
    spray_path = os.path.join(tf2_directory, "tf", "materials", "vgui", "logos", f"{name}.vtf")
    fade_spray.save(spray_path)
    vmt.write_vmt_files(name, tf2_directory)

@cli.command(help = "Create a spray using a static image. Argument is a filepath or a url. Local filepaths are checked before attempting url retrieval.")
@source_argument
@format_option
@aspect_ratio_flag
@name_option
def static(source, format, preserve_aspect_ratio, name):
    tf2_directory = get_tf2_directory()
    images = image_fetching.get_images(source)
    first_frame = images[0]
    name = name or datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    flags =     vtf.TextureFlags.CLAMPS |\
                vtf.TextureFlags.CLAMPT |\
                vtf.TextureFlags.NOLOD |\
                vtf.TextureFlags.EIGHTBITALPHA |\
                vtf.TextureFlags.NOMIP 
    static_spray = vtf.AnimatedSpray([first_frame], vtf.ImageFormats.from_name(format), flags, preserve_aspect_ratio)
    spray_path = os.path.join(tf2_directory, "tf", "materials", "vgui", "logos", f"{name}.vtf")
    static_spray.save(spray_path)
    vmt.write_vmt_files(name, tf2_directory)
    
if __name__ == "__main__":
    cli()


from os.path import join
#TODO: Implement more robust VMT framework for future improvements, for now hardcode strings
#Writes VMTs for spray
def write_vmt_files(file_name: str, tf2_directory: str) -> None:
    logo_dir = join(tf2_directory, "tf", "materials", "vgui", "logos")
    logo_vmt_file = join(logo_dir, f"{file_name}.vmt")
    with open(logo_vmt_file, "wt") as logo_vmt:
        logo_vmt.write(""""UnlitGeneric"
        {{
            "$basetexture"	"vgui/logos/{file_name}"
            "$translucent" "1"
            "$ignorez" "1"
            "$vertexcolor" "1"
            "$vertexalpha" "1"
        }}""".format(file_name = file_name))
    ui_vmt_file = join(logo_dir, "ui", f"{file_name}.vmt")
    with open(ui_vmt_file, "wt") as ui_vmt:
        ui_vmt.write(""""UnlitGeneric"
        {{
            // Original shader: BaseTimesVertexColorAlphaBlendNoOverbright
            "$translucent" 1
            "$basetexture" "VGUI/logos/{file_name}"
            "$vertexcolor" 1
            "$vertexalpha" 1
            "$no_fullbright" 1
            "$ignorez" 1
        }}""".format(file_name = file_name))


#benchmark stuff
from time import time
def benchmark(func): 
    def timedfunc(*args, **kwargs):
        start_time = time()
        ret = func(*args, **kwargs)
        end_time = time()
        print(f"{func.__name__} took: {end_time - start_time} seconds")
        return ret
    return timedfunc


import os.path
import sys

import cv2
import requests
from PIL import Image

#Get url of image source from user input
DUMMY_HEADER = {"User-Agent": "BALLS"}
#Get image(s) from url
#TODO optimize for speed, possibly using ffmpeg
def get_images(path: str, seconds_per_frame: float = 0.2, time_start: float = 0, time_end: float = sys.float_info.max) -> list[Image.Image]:
    #TODO: if switching to ffmpeg, refactor for same function to be used for all animated formats
    def split_animated_image_object(img: Image.Image, seconds_per_frame: float, time_start: float, time_end: float) -> list[Image.Image]:
        extracted_frames = []
        current_time = 0
        frame_start_time = 0
        frame_end_time = 0
        for frame_number in range(img.n_frames):
            img.seek(frame_number)
            frame_duration = img.info['duration'] / 1000 or frame_duration #Last frame duration info sometimes absent for GIFs due to PIL bug
            frame_start_time = frame_end_time
            frame_end_time += frame_duration
            while frame_start_time <= current_time < frame_end_time: 
                if current_time >= time_start:
                    extracted_frames.append(img.copy())
                current_time += seconds_per_frame
                if current_time > time_end: 
                    return extracted_frames
        return extracted_frames
    def split_path_using_cv2(url: str, seconds_per_frame: float, time_start: float, time_end: float) -> list[Image.Image]:
        capture = cv2.VideoCapture(url)
        images = []
        current_time = 0
        frame_start_time = 0
        frame_end_time = 0
        frame_duration = 1 / capture.get(cv2.CAP_PROP_FPS)

        while True:
            capture.grab()
            frame_start_time = frame_end_time
            frame_end_time += frame_duration
            if frame_start_time <= current_time < frame_end_time:
                success, frame = capture.retrieve()
                if not success: break
                while frame_start_time <= current_time < frame_end_time:
                    if current_time >= time_start:
                        image = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB), mode="RGB")
                        images.append(image)
                    current_time += seconds_per_frame
                    if current_time > time_end: 
                        return images 
        return images
    def open_path(path: str, is_local_file: bool) -> Image.Image:
        if is_local_file:
            img = Image.open(path)
        else:
            with requests.get(path, stream = True, headers = DUMMY_HEADER) as response:
                img = Image.open(response.raw)
        return img
    if os.path.isfile(path): #Get resource from local file
        is_local_file = True
        _, ext = os.path.splitext(path)
        file_ending = ext.split(".")[-1]
    else:
        is_local_file = False
        file_ending = path.split("?")[0].split(".")[-1]
    file_ending = file_ending.lower()
    match file_ending:
        case "jpg" | "jpeg":
            img = open_path(path, is_local_file)
            return [img]
        case "png":
            png = open_path(path, is_local_file)
            if png.is_animated:
                return split_animated_image_object(png, seconds_per_frame, time_start, time_end)
            else:
                return [png]
        case "gif":
            gif = open_path(path, is_local_file)
            return split_animated_image_object(gif, seconds_per_frame, time_start, time_end)
        case "webm" | "mp4":
            return split_path_using_cv2(path, seconds_per_frame, time_start, time_end)
        case _:
            raise ValueError(f"Unknown data type, file ending '{file_ending}'")


from dataclasses import dataclass
import enum
import functools
import math
import struct
from typing import ClassVar

import numpy
from PIL import Image
from quicktex import dds, s3tc

class ImageFormats(enum.Enum):
    def __new__(cls, value: int, bits_per_pixel: int, block_side_length: int, implemented: bool):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.bits_per_pixel = bits_per_pixel
        obj.block_side_length = block_side_length
        obj.implemented = implemented
        return obj
    @classmethod
    def from_name(cls, name):
        name = name.upper()
        for member in cls.__members__.values():
            if member.name == name:
                return member
        raise ValueError(f"No enum member with name {name}")
    
    
    NONE = -1, None, None, False
    RGBA8888 = 0, 32, 1, True
    ABGR8888 = 1, 32, 1, True
    RGB888 = 2, 24, 1, True
    BGR888 = 3, 24, 1, True
    RGB565 = 4, 16, 1, False
    I8 = 5, 8, 1, True
    IA88 = 6, 16, 1, True
    P8 = 7, 8, 1, False
    A8 = 8, 8, 1, True
    RGB888_BLUESCREEN = 9, 24, 1, True
    BGR888_BLUESCREEN = 10, 24, 1, True
    ARGB8888 = 11, 32, 1, True
    BGRA8888 = 12, 32, 1, True
    DXT1 = 13, 4, 4, True
    DXT3 = 14, 8, 4, True
    DXT5 = 15, 8, 4, True
    BGRX8888 = 16, 32, 1, True
    BGR565 = 17, 16, 1, True
    BGRX5551 = 18, 16, 1, True
    BGRA4444 = 19, 16, 1, True
    DXT1_ONEBITALPHA = 20, 4, 4, False
    BGRA5551 = 21, 16, 1, True
    UV88 = 22, 16, 1, False
    UVWQ8888 = 23, 32, 1, False
    RGBA16161616F = 24, 64, 1, False
    RGBA16161616 = 25, 64, 1, False
    UVLX8888 = 26, 32, 1, False
class TextureFlags(enum.Flag):
    POINTSAMPLE = 0x00000001
    TRILINEAR = 0x00000002
    CLAMPS = 0x00000004
    CLAMPT = 0x00000008
    ANISOTROPIC = 0x00000010
    HINT_DXT5 = 0x00000020
    PWL_CORRECTED = 0x00000040
    NORMAL = 0x00000080
    NOMIP = 0x00000100
    NOLOD = 0x00000200
    ALL_MIPS = 0x00000400
    PROCEDURAL = 0x00000800

    ONEBITALPHA = 0x00001000
    EIGHTBITALPHA = 0x00002000

    ENVMAP = 0x00004000
    RENDERTARGET = 0x00008000
    DEPTHRENDERTARGET = 0x00010000
    NODEBUGOVERRIDE = 0x00020000
    SINGLECOPY = 0x00040000
    PRE_SRGB = 0x00080000

    UNUSED_00100000 = 0x00100000
    UNUSED_00200000 = 0x00200000
    UNUSED_00400000 = 0x00400000

    NODEPTHBUFFER = 0x00800000

    UNUSED_01000000 = 0x01000000

    CLAMPU = 0x02000000
    VERTEXTEXTURE = 0x04000000
    SSBUMP = 0x08000000            

    UNUSED_10000000 = 0x10000000

    BORDER = 0x20000000

    UNUSED_40000000 = 0x40000000
    UNUSED_80000000 = 0x80000000

class VTFFile:
    @dataclass(kw_only=True)
    class VTFHeader:
        VTF_HEADER_BYTES: ClassVar[int] = 64

        secret_message: bytes = b"YIFFY:3c"

        major_version: int = 7
        minor_version: int = 1
        header_size: int = VTF_HEADER_BYTES
        flags: TextureFlags
        high_res_width: int = 0
        high_res_height: int = 0
        high_res_format: ImageFormats = ImageFormats.DXT1
        low_res_width: int = 0
        low_res_height: int = 0
        low_res_format: ImageFormats = ImageFormats.DXT1
        num_frames: int = 1
        first_frame: int = 0
        num_mipmaps: int = 1
        reflectivity_vector: tuple[int, int, int] = (0, 0, 0)
        bumpmap_scale: int = 0
        def tobytes(self) -> bytes:
            return struct.pack(
            "=4s2IIHHIHH4s3f4sfiBiBBx", #format string
            b"VTF\0", #signature
            self.major_version,
            self.minor_version,
            self.header_size,
            self.high_res_width, 
            self.high_res_height,
            self.flags.value,
            self.num_frames,
            self.first_frame,
            self.secret_message[0:4], #Normally padding, encoding secret message
            *self.reflectivity_vector, #unpack reflectivity vector
            self.secret_message[4:8], #Normally padding, encoding secret message
            self.bumpmap_scale,
            self.high_res_format.value,
            self.num_mipmaps,
            self.low_res_format.value,
            self.low_res_width,
            self.low_res_height,
        ) 
    MAX_BYTES = 512 * 1024
    def __init__(self, **kwargs) -> None:
        self.header = self.VTFHeader(**kwargs)
        self.high_res_data = numpy.empty((self.header.num_mipmaps, self.header.num_frames), dtype='object')
        self.low_res_data = Image.new("RGBA", (0,0))

    @staticmethod
    def get_raw_image_data(img: Image.Image, format: ImageFormats) -> bytes:
        def reorder_channels(img: Image.Image, order: str) -> bytes:
            channels = img.split()
            band_names = img.getbands()
            indices = tuple(band_names.index(channel) for channel in order)
            ordered_channels = tuple(channels[index] for index in indices)
            reordered_image = Image.merge(img.mode, ordered_channels)
            for channel in ordered_channels:
                channel.close()
            return reordered_image
        #NOTE: FIRST channel is stored in the LEAST significant bits
        def trim_channels_uint16(img: Image.Image, channel_depths: tuple[int, ...]):
            assert sum(channel_depths) == 16, "channel depths do not sum to 16 bits!"
            eight_bit_color_array = numpy.asarray(img)
            #Bitshift each channel such that all most significant bits fit within bit depth
            trimmed_array = numpy.right_shift(eight_bit_color_array, [8 - channel_depth for channel_depth in channel_depths]).astype(numpy.uint16)
            #Take bitwise or of each channel, shifted by appropriate amount
            trimmed_data = functools.reduce(lambda x, y: x | y, (trimmed_array[:, :, idx] << sum(channel_depths[:idx]) for idx in range(len(channel_depths))))
            return trimmed_data.tobytes()     
        
        match format:
            case ImageFormats.RGBA8888: #Tested, works
                with img.convert("RGBA") as converted:
                    return converted.tobytes()
            case ImageFormats.ABGR8888: #Tested, works
                with img.convert("RGBA") as converted, reorder_channels(converted, "ABGR") as reordered:
                    return reordered.tobytes()
            case ImageFormats.RGB888: #Tested, works
                with img.convert("RGB") as converted:
                    return converted.tobytes()
            case ImageFormats.BGR888: #Tested, works
                with img.convert("RGB") as converted, reorder_channels(converted, "BGR") as reordered:
                    return reordered.tobytes()
            case ImageFormats.RGB565:
                with img.convert("RGB") as converted:
                    #Not working, might be error with source engine?
                    #return trim_channels_uint16(converted, (5, 6, 5))
                    raise ValueError("DXT1_ONEBITALPHA does not work, use DXT1 instead")
            case ImageFormats.I8: #Tested, works
                with img.convert("L") as converted:
                    return converted.tobytes()
            case ImageFormats.IA88: #Tested, works
                with img.convert("L") as greyscale, img.convert("RGBA") as rgba:
                    greyscale_data = greyscale.tobytes()
                    channels = rgba.split() 
                    alpha_data = channels[-1].tobytes()
                    for channel in channels:
                        channel.close()
                    return b"".join(byte.to_bytes() for pixel in zip(greyscale_data, alpha_data) for byte in pixel)
            case ImageFormats.P8:
                #TODO: Not working, find docs for vtf paletted mode 
                #NOTE: Unsupported by VTFEdit? Might not be able to implement
                with img.convert("P") as converted:
                    palette_data = b"".join(value.to_bytes() for value in converted.getpalette())
                    color_data = converted.tobytes()
                    return palette_data + color_data
            case ImageFormats.A8: #Tested, works
                with img.convert("RGBA") as converted:
                    channels = converted.split()
                    alpha_data = channels[-1].tobytes()
                    for channel in channels:
                        channel.close()
                    return alpha_data
            case ImageFormats.RGB888_BLUESCREEN: #Tested, works
                with img.convert("RGB") as converted:
                    return converted.tobytes()
            case ImageFormats.BGR888_BLUESCREEN: #Tested, works
                with img.convert("RGB") as converted, reorder_channels(converted, "BGR") as reordered:
                    return reordered.tobytes()
            case ImageFormats.ARGB8888: #Tested, works
                with img.convert("RGBA") as converted, reorder_channels(converted, "ARGB") as reordered:
                    return reordered.tobytes()
            case ImageFormats.BGRA8888: #Tested, works
                with img.convert("RGBA") as converted, reorder_channels(converted, "BGRA") as reordered:
                    return reordered.tobytes()
            #TODO: Switch to library with support for DXT1a encoding
            case ImageFormats.DXT1: #Tested, works
                with img.convert("RGBA") as converted:
                    encoder = s3tc.bc1.BC1Encoder()
                    dds_file = dds.encode(converted, encoder, "DXT1", 1)
                    ret = dds_file.textures[0].tobytes()
            case ImageFormats.DXT3: #Tested, works
                with img.convert("RGBA") as converted:
                    encoder = s3tc.bc1.BC1Encoder()
                    dds_file = dds.encode(converted, encoder, "DXT1", 1)
                    color_data = dds_file.textures[0].tobytes()
                    #TODO: Switch to library with built-in BC2 encoding. Current solution is to extract and organize alpha data via bit manipulation
                    pixel_array = numpy.asarray(converted)  
                    
                    alpha_array = pixel_array[:, :, 3]
                    reordered_alpha_data = numpy.concatenate([alpha_array[y:y+4, x:x+4] 
                                                            for y in range(0, alpha_array.shape[0], 4) 
                                                                for x in range(0, alpha_array.shape[1], 4)]).ravel()
                    high_alpha_bits = (reordered_alpha_data[0::2] >> 4) << 4
                    low_alpha_bits = reordered_alpha_data[1::2] >> 4
                    alpha_data = (high_alpha_bits | low_alpha_bits).tobytes()

                    return b"".join(alpha_data[block:block + 8] + color_data[block:block + 8] for block in range(0, len(color_data), 8))
            case ImageFormats.DXT5: #Tested, works
                with img.convert("RGBA") as converted:
                    encoder = s3tc.bc3.BC3Encoder()
                    dds_file = dds.encode(converted, encoder, "DXT5", 1)
                    return dds_file.textures[0].tobytes()
            case ImageFormats.BGRX8888: #Tested, works
                #Alpha data ignored by game engine, assumed to be 255 universally
                with img.convert("RGBA") as converted, reorder_channels(converted, "BGRA") as reordered:
                    return reordered.tobytes()
            #NOTE: It looks like for some reason when dealing with 16 bit formats, the first channel is stored in the least significant bits?
            #NOTE: trim_channels_uint16 function accounts for this, but it is a weird implementation
            case ImageFormats.BGR565: #Tested, works
                with img.convert("RGB") as converted, reorder_channels(img, "BGR") as reordered:
                    return trim_channels_uint16(reordered, (5, 6, 5))       
            case ImageFormats.BGRX5551: #Tested, works
                #Alpha data ignored by game engine, assumed to be 1 universally
                with img.convert("RGBA") as converted, reorder_channels(converted, "BGRA") as reordered:
                    return trim_channels_uint16(reordered, (5, 5, 5, 1))   
            case ImageFormats.BGRA4444: #Tested, works
                with img.convert("RGBA") as converted, reorder_channels(converted, "BGRA") as reordered:
                    return trim_channels_uint16(reordered, (4, 4, 4, 4))
            case ImageFormats.DXT1_ONEBITALPHA: #Broken in source
                raise ValueError("DXT1_ONEBITALPHA does not work, use DXT1 instead")
            case ImageFormats.BGRA5551: #Tested, works
                with img.convert("RGBA") as converted, reorder_channels(converted, "BGRA") as reordered:
                    return trim_channels_uint16(reordered, (5, 5, 5, 1))
            case _:
                raise ValueError(f"Unimplemented Data Format: {format}")
        return ret
    @staticmethod
    def resize_image(image: Image.Image, side_length: int, preserve_aspect_ratio: bool):
        if preserve_aspect_ratio:
            if image.width > image.height:
                w, h = side_length, round(side_length * (image.height / image.width))
            else:
                w, h = round(side_length * (image.width / image.height)), side_length
            with image.resize((w, h)) as resized_image: 
                x_offset = (side_length - w) // 2
                y_offset = (side_length - h) // 2
                result = Image.new("RGBA", (side_length, side_length), (0, 0, 0, 0))
                result.paste(resized_image, (x_offset, y_offset))
        else:
            result = image.resize((side_length, side_length))
        return result
    
    def assert_full(self):
        assert None not in self.high_res_data, "All mipmaps and frames must be filled!"
    def tobytes(self) -> bytes:
        self.assert_full()
        header_bytes = [self.header.tobytes()]
        low_res_bytes = [self.get_raw_image_data(self.low_res_data, self.header.low_res_format)] if 0 not in self.low_res_data.size else []
        high_res_bytes = [self.get_raw_image_data(img, self.header.high_res_format) for img in self.high_res_data.ravel()]
        file_data = b''.join(header_bytes + low_res_bytes + high_res_bytes)
        return file_data
    def save(self, path) -> None:
        with open(path, "wb") as file:
            file.write(self.tobytes())
    def calculate_max_high_res_bits(self) -> int:
        VTF_HEADER_BITS = self.header.VTF_HEADER_BYTES * 8
        MAX_VTF_BITS = self.MAX_BYTES * 8

        low_res_bits = self.header.low_res_height * self.header.low_res_width * self.header.low_res_format.bits_per_pixel
        max_high_res_bits = MAX_VTF_BITS - VTF_HEADER_BITS - low_res_bits
        return max_high_res_bits
        
    def normalize_mipmap_sizes(self) -> None:
        self.assert_full()
        #Iterate over mipmaps in order of decreasing size
        for idx, mipmap in enumerate(self.high_res_data[::-1]):
            resize_factor = 2 ** idx
            mipmap_width = max(self.header.high_res_width // resize_factor,  1)
            mipmap_height = max(self.header.high_res_height // resize_factor, 1)
            for idx, frame in enumerate(mipmap):
                mipmap[idx] = frame.resize((mipmap_width, mipmap_height))

class FadeSpray(VTFFile):
    SMALLEST_MIPMAP_SIDE_LENGTH = 32
    def calculate_num_mipmaps(self) -> int:
        MIPMAP_PIXEL_RATIO = 4
 
        max_high_res_bits = self.calculate_max_high_res_bits()
        max_high_res_pixels = max_high_res_bits // self.header.high_res_format.bits_per_pixel

        max_high_res_pixels //= self.header.num_frames

        #Mipmap pixel sum is defined by a geometric series, solve for maximum number of mipmaps that will still fit within 512 KB VTF file
        num_mipmaps = math.floor(math.log(1 - ((max_high_res_pixels * (1 - MIPMAP_PIXEL_RATIO)) / (self.SMALLEST_MIPMAP_SIDE_LENGTH ** 2)), MIPMAP_PIXEL_RATIO))
        return num_mipmaps
    def __init__(self, images: Image.Image, format: ImageFormats, flags: TextureFlags, preserve_aspect_ratio: bool):
        #Base VTFFile must be inited before calculating minimum mipmap sizes
        VTFFile.__init__(self, high_res_format = format, flags = flags)
        self.header.num_mipmaps = self.calculate_num_mipmaps()
        assert self.header.num_mipmaps >= len(images), "Too many images for faded spray!"
        self.header.high_res_width = self.header.high_res_height = self.SMALLEST_MIPMAP_SIDE_LENGTH * 2 ** (self.header.num_mipmaps - 1)
        #Resize high res data array, originally sized for one mipmap
        self.high_res_data = numpy.empty((self.header.num_mipmaps, self.header.num_frames), dtype='object')
        for idx, image in enumerate(reversed(images)):
            self.high_res_data[-(1 + idx), :] = [image for _ in range(self.header.num_frames)]
        #If mipmaps are activated then the game will attempt to read all mipmaps down the minumum. Smallest mipmap will be used for all smaller mipmaps to avoid reading garbage data
        for mipmap in self.high_res_data:
            for idx, frame in enumerate(mipmap):
                if frame is None:
                    mipmap[idx] = images[0]
        
        
        for idx, mipmap in enumerate(self.high_res_data):
            mipmap_side_length = self.SMALLEST_MIPMAP_SIDE_LENGTH * (2 ** idx)
            for idx2, frame in enumerate(mipmap):
                mipmap[idx2] = self.resize_image(frame, mipmap_side_length, preserve_aspect_ratio)
            


class AnimatedSpray(VTFFile):
    def calculate_frame_side_length(self) -> int:
        max_high_res_bits = self.calculate_max_high_res_bits()
        max_bits_per_frame = max_high_res_bits // self.header.num_frames
        max_pixels_per_frame = max_bits_per_frame // self.header.high_res_format.bits_per_pixel

        max_possible_side_length = math.floor(math.sqrt(max_pixels_per_frame))
        actual_side_length = (max_possible_side_length // self.header.high_res_format.bits_per_pixel) * self.header.high_res_format.bits_per_pixel
        return actual_side_length
        
    def __init__(self, images: list[Image.Image], format: ImageFormats, flags: TextureFlags, preserve_aspect_ratio: bool):
        VTFFile.__init__(self, high_res_format = format, num_frames = len(images), flags = flags)
        self.header.high_res_height = self.header.high_res_width = side_length = self.calculate_frame_side_length()
        for idx, image in enumerate(images):
            images[idx] = self.resize_image(image, side_length, preserve_aspect_ratio)
        self.high_res_data[0, :] = images
